from abc import ABC, abstractmethod
from typing import List, TypeVar, Generic
import re
import unittest
from unitpy.Test import Test

T = TypeVar('T')
class GeneralTests ( Generic[T], ABC):
    '''
    class taht runs tests on hashCode, equals, toString.
    
    To use this:
    <ol>
    <li>Let your junit test extend this
    <li>Create a function annnotated @DataPoints("equal") that returns a list of
    type T with objects that are all equal to each other. There should be at
    least 2 of these that were constructed with the same constructor but
    separately, so that we can check if two objects that were constructed
    identically are actually equal.
    <li>Create a function annotated with @DataPoints("notequal") that returns a
    list of objects of type T that are all not equal to those in the "equal" set.
    <li>Crate a static String annotated with @DataPoint("equalstring") that
    contains a java regex that should match with the result of calling toString
    on the objects from the "equal" datapoints.
    
    </ol>
    NOTICE this is NOT a junit test, and it will fail if you run this from
    Eclipse.
    
    @param [T] the type of the object under test.
    '''
    STRINGSNONULLELEMENTS = "getGeneralTestStrings must not contain null elements in the list"
    ATLEAST1 = "each equallists from getGeneralTestStrings must contain at least one element"
    NONULLELEMENTS = "getGeneralTestData must not contain null elements in the list"
    LISTSEQUALSIZE = "getGeneralTestData and getGeneralTestStrings must return list of same size."
    DATAATLEAST2 = "getGeneralTestData must return list of least 2 elements"
    STRINGATLEAST2 = "getGeneralTestStrings must return list of least 2 elements"

    # for testing un-equality.
    class Dummy:
        pass

    dummy = Dummy()

    @abstractmethod
    def getGeneralTestData(self) -> List[List[T]]:
        '''
        @return a list containing equallists of the terms T under test. The outer
                list must contain at least 2 equallists. The equallists are lists
                 of identical terms that are supposed to return true on their
                .equals function. All elements in equallist must contain at least
                1 term. At least one equallist must contain more than 1 term.
        '''

    @abstractmethod
    def getGeneralTestStrings(self) -> List[str]:
        '''
        @return a list of equal length as the (top level) list returned by
                {@link #getGeneralTestData()}. It contains a regexp that should
                match the toString of the terms in the equallists.
        '''
#     @Rule
#     public ExpectedException expectedException = ExpectedException.none();

    # Cached copies of the data. Use names that are unlikely to collide with
    # actual tests.
    _data_: List[List[T]] =None  #type:ignore
    _strings_: List[str] =None  #type:ignore

    @Test()
    def testValidateData(self):
        assertTrue=unittest.TestCase().assertTrue
        assertFalse=unittest.TestCase().assertFalse
        assertEqual=unittest.TestCase().assertEqual
        
        self._getTheData()
        assertFalse(self._data_==None, self.DATAATLEAST2)
        assertFalse(self._strings_==None, self.STRINGATLEAST2 )
        assertTrue(len(self._data_) >= 2, self.DATAATLEAST2 )
        assertEqual(len(self._strings_), len(self._data_),self.LISTSEQUALSIZE)
        for d  in self. _data_:
            assertFalse(d==None, self.NONULLELEMENTS)
            assertTrue( len(d) > 0,self.ATLEAST1)
        
        for string in self._strings_:
            assertFalse( string==None, self.STRINGSNONULLELEMENTS )

        themax =0;
        for d in self._data_:
            themax = max(len(d),themax)
        assertTrue(themax>=2, \
            "At least one equallist in getGeneralTestData must contain st least 2 elements")

    @Test()
    def testToStringAllData(self):
        self._getTheData();
        for  n in range(len(self._data_)):
            for point in self._data_[n]:
                self._toStringTest(point, self._strings_[n])


    @Test()
    def testEqualsAllData(self):
        self._getTheData();
        for equalset in self._data_:
            for value1 in equalset:
                for value2 in equalset:
                    self._testEquals(value1, value2)


    @Test()
    def testNotEqualsAllData(self):
        self._getTheData();
        for n in range(len(self._data_)):
            for m in range(len(self._data_)):
                if (n == m):
                    continue
                for value1 in self._data_[n]:
                    for value2 in self._data_[m]:
                        self._testNotEquals(value1, value2);


    def _getTheData(self):
        '''
        Ensure we have the data loaded in our private fields, to avoid creating
        it repeatedly. The clients may have to do expensive computations to make
        the test data.
        '''
        if not self._data_:
            self._data_ = self.getGeneralTestData()

        if not self._strings_:
            self._strings_ = self.getGeneralTestStrings()


    def _toStringTest(self, value:object, pattern:str):
        '''
        Check that value.toString() matches expected string.
        
        @param value
        @param pattern the pattern that value is supposed to match.
        '''
        assertTrue=unittest.TestCase().assertTrue
        assertFalse=unittest.TestCase().assertFalse

        assertFalse( value==None, "Encountered a null object in your testset")
        assertFalse( pattern==None, "Encountered a null string in your testset")
        assertFalse(repr(value)==None, "repr is returning None!" )
        assertTrue(re.fullmatch(pattern, repr(value))!=None, \
                "Pattern "+pattern+" does not match string "+repr(value))

    def _testEquals(self, value1:T, value2:T):
        assertTrue=unittest.TestCase().assertTrue
        assertFalse=unittest.TestCase().assertFalse


        err = "values <" + str(value1) + "> and <" + str(value2)\
                + "> should be equal but are not."
        errn = "values <" + str(value1) + "> and <" + str(value2)\
                + "> should have equal hashcode.";
        # because both value1 and value2 iterate over the whole set, this also
        # checks symmetry.
        assertFalse(value1==None,"value " + str(value1) + " is null");
        assertFalse(value2==None, "value " + str(value2) + " is null");
        assertTrue( value1==value2, err)
        assertTrue(hash(value1) == hash(value2), errn)
        assertFalse( value1==None, err);
        # we test with all values for value1 and value2 anyway so test only
        # value1
        assertFalse(value1==self.dummy, \
            "value " + str(value1) + "should be unequal to dummy object");

    def _testNotEquals(self, value1:T , value2:T ):
        assertFalse=unittest.TestCase().assertFalse
        assertNotEqual=unittest.TestCase().assertNotEqual

        err = "values <" + str(value1) + "< and <" + str(value2) \
                + "> should be unequal but are equal."
        errh = "Warning: values <" + str(value1) + "< and <" + str(value2) \
                + "> are different but actually have same hashcode. This may be incidental but then please use slightly different test values";
        # we call equals() explicitly (9nstead of assertEquals etc) to make
        # sure what  we call
        assertFalse( value1==None,\
                "encountered null value in list of objects to test for equality" )
        assertFalse(value2==None, \
                "encountered null value in list of objects to test for equality" )
        assertFalse(value1==value2 ,err )
        assertFalse(value2==value1,err)
        assertNotEqual(hash(value1),hash(value2),errh)
        assertFalse(value2==None,err)
        assertNotEqual(value1, "*",err)
        assertNotEqual(value2, "*",err)
